"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.visitStatementBase = exports.visitor = void 0;
exports.visitor = {
    visitUrichk(visitor, node) {
        for (const rule of node) {
            visitor.visitRule(visitor, rule);
        }
    },
    visitRule(visitor, node) {
        visitStatementBase(visitor, node, () => {
            visitor.visitHead(visitor, node.head);
            visitor.visitTail(visitor, node.tail);
        });
    },
    visitHead(visitor, node) {
        visitStatementBase(visitor, node, () => {
            const { scheme, authority, path } = node;
            if (scheme)
                visitor.visitScheme(visitor, scheme);
            if (authority)
                visitor.visitAuthority(visitor, authority);
            if (path)
                visitor.visitPath(visitor, path);
        });
    },
    visitTail(visitor, node) {
        visitStatementBase(visitor, node, () => {
            visitor.visitToken(visitor, node.openParens);
            for (const rule of node.rules) {
                visitor.visitTailRule(visitor, rule);
            }
            visitor.visitToken(visitor, node.closeParens);
        });
    },
    visitScheme(visitor, node) {
        visitor.visitToken(visitor, node);
    },
    visitAuthority(visitor, node) {
        const { userinfo, host, port } = node;
        if (userinfo)
            visitor.visitUserinfo(visitor, userinfo);
        visitor.visitHost(visitor, host);
        if (port)
            visitor.visitPort(visitor, port);
    },
    visitUserinfo(visitor, node) {
        visitor.visitToken(visitor, node);
    },
    visitHost(visitor, node) {
        visitor.visitToken(visitor, node);
    },
    visitPort(visitor, node) {
        visitor.visitToken(visitor, node);
    },
    visitPath(visitor, node) {
        for (const pathFragment of node.fragments) {
            visitor.visitPathFragment(visitor, pathFragment);
        }
    },
    visitPathFragment(visitor, node) {
        if (node.type === "static") {
            visitor.visitStaticPathFragment(visitor, node);
        }
        else if (node.type === "param") {
            visitor.visitParamPathFragment(visitor, node);
        }
    },
    visitStaticPathFragment(visitor, node) {
        visitor.visitToken(visitor, node);
    },
    visitParamPathFragment(visitor, node) {
        visitor.visitToken(visitor, node);
    },
    visitTailRule(visitor, node) {
        if (isTailRuleMatch(node))
            visitor.visitTailRuleMatch(visitor, node);
        if (isTailRuleForm(node))
            visitor.visitTailRuleForm(visitor, node);
        function isTailRuleMatch(node) {
            return node.matchType.text === "match";
        }
        function isTailRuleForm(node) {
            return node.matchType.text === "form";
        }
    },
    visitTailRuleMatch(visitor, node) {
        visitStatementBase(visitor, node, () => {
            const { tailType, matchType, pattern } = node;
            visitor.visitTailType(visitor, tailType);
            visitor.visitMatchType(visitor, matchType);
            for (const value of pattern) {
                visitor.visitTailRulePatternValue(visitor, value);
            }
        });
    },
    visitTailRulePatternValue(visitor, node) {
        visitStatementBase(visitor, node, () => {
            visitor.visitToken(visitor, node);
        });
    },
    visitTailRuleForm(visitor, node) {
        visitStatementBase(visitor, node, () => {
            const { tailType, matchType, pattern } = node;
            visitor.visitTailType(visitor, tailType);
            visitor.visitMatchType(visitor, matchType);
            for (const rule of pattern) {
                visitor.visitTailRuleFormPatternRule(visitor, rule);
            }
        });
    },
    visitTailRuleFormPatternRule(visitor, node) {
        visitStatementBase(visitor, node, () => {
            const { key, value, array } = node;
            visitor.visitKey(visitor, key);
            if (array)
                visitor.visitArrayToken(visitor, array);
            if (value) {
                for (const patternValue of value) {
                    visitor.visitTailRulePatternValue(visitor, patternValue);
                }
            }
        });
    },
    visitKey(visitor, node) {
        visitor.visitToken(visitor, node);
    },
    visitArrayToken(visitor, node) {
        visitor.visitToken(visitor, node);
    },
    visitTailType(visitor, node) {
        visitor.visitToken(visitor, node);
    },
    visitMatchType(visitor, node) {
        visitor.visitToken(visitor, node);
    },
    visitCommentGroup(visitor, node) {
        for (const comment of node.comments) {
            visitor.visitComment(visitor, comment);
        }
    },
    visitComment(visitor, node) {
        visitor.visitToken(visitor, node);
    },
    visitToken() { },
};
function visitStatementBase(visitor, node, visit) {
    for (const commentGroup of node.leadingComments) {
        visitor.visitCommentGroup(visitor, commentGroup);
    }
    visit();
    for (const commentGroup of node.trailingComments) {
        visitor.visitCommentGroup(visitor, commentGroup);
    }
}
exports.visitStatementBase = visitStatementBase;
